/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.implementation.exception.InvalidPropertyException;
import org.ow2.petals.microkernel.api.implementation.exception.MissingPropertyException;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.system.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.system.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.registry.overlay.api.RegistryOverlayAPI;
import org.ow2.petals.microkernel.registry.overlay.exception.AuthenticationException;
import org.ow2.petals.microkernel.registry.overlay.exception.InvalidConfigurationException;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.api.RegistryEndpointFactory;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint.Type;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationBuilder;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationException;
import org.ow2.petals.registry_overlay.client.configuration.generated.Configuration;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMember;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMembers;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.Topology;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;
import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.client.config.ClientConfig;
import com.hazelcast.client.config.ClientNetworkConfig;
import com.hazelcast.config.NearCacheConfig;
import com.hazelcast.config.SerializationConfig;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.ILock;
import com.hazelcast.core.IMap;
import com.hazelcast.query.Predicates;
import com.hazelcast.query.SqlPredicate;

/**
 * The registry implementation based on the Petals Registry Overlay.
 * 
 * TODO endpoints should be stored by serviceName AND endpointName! the JBI API is not clear on that but it seems that
 * it would make sense since an endpointName is NEVER used alone, always accompanied by a serviceName.
 * 
 * Note: some of these methods (see e.g., {@link #getEndpoints(EndpointType)}) return lazy collection (aka live view)
 * but they never relies on hazelcast: once the collection is returned, no hazelcast requests are done!
 * 
 * TODO make the live-view not recompute their value everytime they are called!
 * 
 * @author Christophe DENEUX - Linagora
 * @author Victor Noel
 */

@Component(provides = @Interface(name = SharedAreaService.FRACTAL_SRV_ITF_NAME, signature = SharedAreaService.class))
public class RegistryOverlayImpl implements RegistryOverlayAPI {

    /**
     * Prefix of property names
     */
    private static final String PROPERTY_NAME_PREFIX = RegistryOverlayImpl.class.getName() + ".";
    
    /**
     * Prefix of property names configuring map cache
     */
    private static final String PROPERTY_NAME_MAP_CACHE_PREFIX = PROPERTY_NAME_PREFIX + "map-cache.";

    /**
     * Name of the property defining the max size of map cache, in endpoint number
     */
    public static final String PROPERTY_NAME_MAP_CACHE_MAX_SIZE = PROPERTY_NAME_MAP_CACHE_PREFIX + "max-size";

    /**
     * Name of the property associated to the group name
     */
    private static final String GROUP_NAME_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_NAME;
    
    /**
     * Name of the property associated to the group password
     */
    private static final String GROUP_PASSWORD_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD;

    /**
     * Name pattern of the property associated to an overlay member host name
     */
    private static final String MEMBER_HOST_PROPERTY_NAME_PATTERN = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.ADDRESS + ".<x>";

    /**
     * Default value of the property {@value #PROPERTY_NAME_MAP_CACHE_MAX_SIZE}
     */
    public static final int DEFAULT_MAP_CACHE_MAX_SIZE = 500;

    /**
     * The map containing endpoints.
     */
    private IMap<String, StoredRegistryEndpoint> endpointsMap;

    /**
     * <p>
     * The map containing endpoint description.
     * </p>
     * <p>
     * To avoid classloader problem, the WSDL is stored as {@link String}.
     * </p>
     */
    private IMap<String, String> endpointDescriptionsMap;

    /**
     * The distributed lock managing access to the topology object
     */
    private ILock topologyObjectLock;

    /**
     * <p>
     * The map containing dynamic topologies.
     * </p>
     * <p>
     * Note: Type is {@link IMap} to be able to register a listener on updated entries.
     * </p>
     */
    private IMap<String, Topology> topologiesMap;

    /**
     * Identifier of the listener on the topology entries
     */
    private String dynamicTopologyListener;

    /**
     * Configuration service fractal component
     */
    @Requires(name = SharedAreaService.CONFIGURATION_FRACTAL_ITF_NAME)
    protected ConfigurationService configurationService;

    /**
     * Logging utilities
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(COMPONENT_LOGGER_NAME));

    /**
     * The topology name associated to the current domain in the registry
     */
    private String topologyName;

    /**
     * The Hazelcast client
     */
    private HazelcastInstance hzlClient;

    private Configuration registryOverlayConfiguration = null;

    @Override
    public Collection<PetalsServiceEndpoint> getEndpoints(final EndpointType type) throws SharedAreaException {
        try {
            // values() returns a copy of the values
            final Collection<StoredRegistryEndpoint> values = type == null ? this.endpointsMap.values()
                    : this.endpointsMap.values(Predicates.equal("type", toStoredType(type)));
            return Collections2.transform(values, STORED_2_PETALS);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    /**
     * TODO See {@link ExternalEndpointImpl} for PETALSESBCONT-393
     */
    @Override
    public PetalsServiceEndpoint registerExternalEndpoint(final ServiceEndpoint endpoint, final Location location,
            final Document description, final boolean canOverwrite) throws SharedAreaException {
        // we pack it so that it can be serialized in the shared area!
        final StoredRegistryEndpoint storedEndpoint = toStoredEndpoint(Type.EXTERNAL, endpoint, location);
        if (registerEndpoint(storedEndpoint, description, canOverwrite)) {
            return new ExternalEndpointImpl(storedEndpoint);
        } else {
            return null;
        }
    }

    @Override
    public PetalsServiceEndpoint registerInternalEndpoint(final String endpointName, final QName serviceName,
            final QName[] interfaces, Location location, Document description, boolean canOverwrite)
                    throws SharedAreaException {

        final StoredRegistryEndpoint storedEndpoint = toStoredEndpoint(Type.INTERNAL, endpointName, serviceName,
                interfaces, location);
        if (registerEndpoint(storedEndpoint, description, canOverwrite)) {
            return new InternalEndpointImpl(storedEndpoint);
        } else {
            return null;
        }
    }

    private boolean registerEndpoint(final StoredRegistryEndpoint serviceEndpoint,
            final Document description, final boolean canOverwrite) throws SharedAreaException {

        if (serviceEndpoint == null) {
            throw new SharedAreaException("Can not register a null service endpoint");
        }

        final String endpointName = serviceEndpoint.getEndpointName();

        try {
            // We register the endpoint
            if (!canOverwrite) {
                if (this.endpointsMap.putIfAbsent(endpointName, serviceEndpoint) != null) {
                    // there was already a value, so the new one wasn't added
                    return false;
                }
            } else {
                // use set because we don't need the previous value
                this.endpointsMap.set(endpointName, serviceEndpoint);
            }

            // For performance reasons, the description is not saved as part of the service endpoint. It is retrieved
            // only if needed through ServiceEndpoint.getDescription(). See #retrieveDescription() used by
            // RegistryOverlayEndpoint.
            if (description != null) {
                try {
                    // we omit declaration, it will save a bit of space...
                    // TODO replace with proper DOM serialization?!
                    this.endpointDescriptionsMap.set(endpointName,
                            XMLHelper.createStringFromDOMNode(description, true));
                } catch (final TransformerException e) {
                    throw new SharedAreaException("Can't serialize the description into the registry", e);
                }
            }

            return true;
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public PetalsServiceEndpoint deregisterEndpoint(final QName serviceName, final String endpointName)
            throws SharedAreaException {
        try {
            
            final StoredRegistryEndpoint stored = this.endpointsMap.remove(endpointName);

            // use delete because we don't need the value
            this.endpointDescriptionsMap.delete(endpointName);
            
            return stored != null ? STORED_2_PETALS.apply(stored) : null;
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public boolean isRegisteredEndpoint(final QName serviceName, final String endpointName) throws SharedAreaException {
        try {
            return this.endpointsMap.containsKey(endpointName);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForInterface(final QName interfaceName,
            final EndpointType type) throws SharedAreaException {

        try {
            // values() returns a copy of the values
            final SqlPredicate snP = new SqlPredicate(
                    String.format("interfaces[any] = '%s'", interfaceName.toString()));
            final Collection<StoredRegistryEndpoint> values = this.endpointsMap
                    .values(type == null ? snP : Predicates.and(Predicates.equal("type", toStoredType(type)), snP));
            return Collections2.transform(values, STORED_2_PETALS);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public PetalsServiceEndpoint getEndpoint(final QName service, final String endpointName)
            throws SharedAreaException {

        assert service != null;
        assert endpointName != null;

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + service.toString() + ", endpointName = " + endpointName);
        }

        try {
            final StoredRegistryEndpoint stored = this.endpointsMap.get(endpointName);
            // We check that the endpoint implements the expected service
            if (stored != null && stored.getServiceName().equals(service.toString())) {
                return STORED_2_PETALS.apply(stored);
            }
            return null;
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        } finally {
            this.log.end();
        }
    }

    private Collection<PetalsServiceEndpoint> getLocalEndpoints(final EndpointType type) throws SharedAreaException {
        this.log.start();

        try {
            final String localContainerName = this.configurationService.getContainerConfiguration().getName();
            return Collections2.filter(getEndpoints(type), new Predicate<PetalsServiceEndpoint>() {
                @Override
                public boolean apply(final PetalsServiceEndpoint endpoint) {
                    return (type == null || type.equals(endpoint.getType()))
                            && endpoint.getLocation().getContainerName().equals(localContainerName);
                }
            });
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    private Collection<SavedEndpoint> removeAllLocalEndpoints() throws SharedAreaException {
        this.log.start();

        try {
            final Collection<PetalsServiceEndpoint> localEndpoints = getLocalEndpoints(null);
            // let's build a collection
            final Collection<SavedEndpoint> endpoints = new ArrayList<>(localEndpoints.size());
            for (final PetalsServiceEndpoint endpoint : localEndpoints) {
                final Document description = toDocument(
                        this.endpointDescriptionsMap.remove(endpoint.getEndpointName()));
                this.endpointsMap.delete(endpoint.getEndpointName());
                endpoints.add(new SavedEndpoint(endpoint, description));
            }
            return endpoints;
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    private void reactivateAllLocalEndpoints(final Collection<SavedEndpoint> endpoints)
            throws SharedAreaException {
        this.log.start();

        try {
            for (final SavedEndpoint endpoint : endpoints) {
                if (endpoint.endpoint.getType() == EndpointType.EXTERNAL) {
                    registerExternalEndpoint(endpoint.endpoint, endpoint.endpoint.getLocation(), endpoint.description,
                            false);
                } else if (endpoint.endpoint.getType() == EndpointType.INTERNAL) {
                    registerInternalEndpoint(endpoint.endpoint.getEndpointName(), endpoint.endpoint.getServiceName(),
                            endpoint.endpoint.getInterfaces(), endpoint.endpoint.getLocation(), endpoint.description,
                            false);
                } else {
                    throw new SharedAreaException("Impossible case");
                }
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Lifecycle(step = Step.START)
    public void start() throws SharedAreaException {
        this.log.start();

        try {
            final String domainName = this.configurationService.getDomainConfiguration().getName();
            if (domainName == null) {
                throw new InvalidConfigurationException("The domain name is missing");
            }
            this.topologyName = domainName;

            final List<Object> extraParameters = this.configurationService.getDomainConfiguration()
                    .getExtraParameters();
            if (extraParameters != null) {
                try {
                    for (final Object extraParameter : extraParameters) {
                        final Configuration cfg = ConfigurationBuilder
                                .createConfiguration((org.w3c.dom.Node) extraParameter);
                        if (cfg != null) {
                            this.registryOverlayConfiguration = cfg;
                            break;
                        }
                    }
                } catch (final ConfigurationException e) {
                    throw new InvalidConfigurationException(e);
                }
                if (this.registryOverlayConfiguration == null) {
                    throw new InvalidConfigurationException("Missing configuration");
                }
            } else {
                throw new InvalidConfigurationException("Missing configuration");
            }

            if (this.registryOverlayConfiguration.getGroupName() == null
                    || this.registryOverlayConfiguration.getGroupName().isEmpty()) {
                throw new InvalidConfigurationException("The group name is missing");
            }
            if (this.registryOverlayConfiguration.getGroupPassword() == null
                    || this.registryOverlayConfiguration.getGroupPassword().isEmpty()) {
                throw new InvalidConfigurationException("The group password is missing");
            }
            if (this.registryOverlayConfiguration.getOverlayMembers() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember().isEmpty()) {
                throw new InvalidConfigurationException("No overlay member declared");
            }

            if (this.log.isConfigEnabled()) {
                this.log.config("Registry name: " + this.topologyName);
                this.log.config("Group name: " + this.registryOverlayConfiguration.getGroupName());
                this.log.config("Group pwd: ****");
            }

            // Connect to the Hazelcast cluster
            final ClientConfig cltCfg = new ClientConfig();

            cltCfg.getGroupConfig().setName(this.registryOverlayConfiguration.getGroupName());
            cltCfg.getGroupConfig().setPassword(this.registryOverlayConfiguration.getGroupPassword());
            final ClientNetworkConfig networkConfig = new ClientNetworkConfig();
            cltCfg.setNetworkConfig(networkConfig);
            for (final OverlayMember overlayMember : this.registryOverlayConfiguration.getOverlayMembers()
                    .getOverlayMember()) {
                final String host = overlayMember.getValue();
                final BigInteger port = overlayMember.getPort();
                if (host == null || host.isEmpty()) {
                    throw new InvalidConfigurationException("The hostname of an overlay member is missing");
                }
                final String address = host + ":" + (port == null ? Constants.DEFAULT_PORT : port.toString());
                if (this.log.isConfigEnabled()) {
                    this.log.config("Member: " + address);
                }
                networkConfig.addAddress(address);
            }

            // Enable near cache
            this.enableNearCache(cltCfg);

            // register the serialization factory
            cltCfg.setSerializationConfig(new SerializationConfig().addDataSerializableFactoryClass(
                    RegistryEndpointFactory.REGISTRY_ENDPOINT_FACTORY_ID, RegistryEndpointFactory.class));

            // We use the Petals ESB container classloader
            cltCfg.setClassLoader(Thread.currentThread().getContextClassLoader());

            try {
                this.hzlClient = HazelcastClient.newHazelcastClient(cltCfg);

                // When no more clients associated to a dedicated topology is connected, all data relative to the
                // topology must be removed from the HZL cluster. To manage nominal client shutdown and unexpected
                // client shutdown, the HZL cluster will detect itself when clients are disconnected. When no more
                // client is connected, the HZL cluster will clean up topology information.
                // That's why we register the client, as <ip-address:port> in a distributed map. The HZL cluster will
                // deregister the client on disconnection event.
                final IMap<String, String> clients = DataModelApi.getClientsMap(this.hzlClient);
                clients.set(this.hzlClient.getLocalEndpoint().getUuid(), this.topologyName);

                // Distributed objects about endpoints
                this.endpointsMap = DataModelApi.getEndpointsMap(this.hzlClient, this.topologyName);
                this.endpointDescriptionsMap = DataModelApi.getEndpointDescriptionsMap(this.hzlClient,
                        this.topologyName);

                // Distributed objects about dynamic topology
                this.topologyObjectLock = DataModelApi.getTopologiesMapLock(this.hzlClient, this.topologyName);

                this.topologiesMap = DataModelApi.getTopologiesMap(this.hzlClient);

            } catch (final RuntimeException e) {
                if (e.getCause() != null && e.getCause() instanceof com.hazelcast.client.AuthenticationException) {
                    throw new AuthenticationException();
                } else {
                    throw new SharedAreaException(e);
                }
            }
        } catch (final SharedAreaException e) {
            throw e;
        } catch (final Exception e) {
            throw new SharedAreaException(e);
        } finally {
            this.log.end();
        }
    }
    
    @Override
    public void loadTopology(final Collection<SavedEndpoint> endpoints) throws SharedAreaException {

        if (this.dynamicTopologyListener != null) {
            throw new SharedAreaException("Can't load the topology if we already attached to one!");
        }

        final int lockWaitTime = this.configurationService.getContainerConfiguration().getTopologyDynamicLockWaitTime();

        try {
            if (this.topologyObjectLock.tryLock(lockWaitTime, TimeUnit.SECONDS)) {
                try {
                    Topology topology = this.topologiesMap.get(this.topologyName);
                    if (topology == null) {
                        // topology is not in the registry, let's use the local one to bootstrap the domain in the
                        // registry
                        topology = this.configurationService.getLocalTopology();
                        this.log.info("First startup of a Petals ESB node of the topology, the dynamic topology"
                                + " is initialized with the local topology definition.");
                        if (this.log.isDebugEnabled()) {
                            this.log.debug("Applied topology: " + topology);
                        }

                        // note: the topology in the ConfigurationService was already validated

                        this.topologiesMap.put(this.topologyName, topology);
                    } else {
                        // A topology exists, we use it
                        if (this.log.isDebugEnabled()) {
                            this.log.debug("Topology used from the registry: " + topology);
                        }

                        // this will validate and update the retrieved topology
                        this.configurationService.updateTopology(topology);
                    }

                    // and finally activate the topology listener for future events
                    // (before releasing the lock, so that nobody can do modification before I register it!)
                    this.dynamicTopologyListener = this.topologiesMap.addEntryListener(
                            new DynamicTopologyListener(this.configurationService, this.topologyName, this.log),
                            this.topologyName, true);
                } finally {
                    this.topologyObjectLock.unlock();
                }

                reactivateAllLocalEndpoints(endpoints);

            } else {
                throw new SharedAreaException(String
                        .format("The lock on the topology object is already acquired (waited %s).", lockWaitTime));
            }
        } catch (org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException e) {
            throw new SharedAreaException("Invalid topology", e);
        } catch (InterruptedException e) {
            throw new SharedAreaException("Locking was interrupted");
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public void joinTopology(final Collection<SavedEndpoint> endpoints) throws SharedAreaException {
        final int lockWaitTime = this.configurationService.getContainerConfiguration().getTopologyDynamicLockWaitTime();

        if (this.dynamicTopologyListener != null) {
            throw new SharedAreaException("Can't join a topology if we didn't leave one before");
        }

        try {
            if (this.topologyObjectLock.tryLock(lockWaitTime, TimeUnit.SECONDS)) {
                try {
                    final Topology topology = this.topologiesMap.get(this.topologyName);
                    if (topology == null) {
                        throw new SharedAreaException("Can't find the topology in the registry!");
                    }

                    final ContainerConfiguration currentContainerConfig = this.configurationService
                            .getContainerConfiguration();

                    // if it's a dynamic topology and we are not in it, let's add ourselves!
                    if (topology.getDomain().getMode() == DomainMode.DYNAMIC
                            && TopologyHelper.findContainer(currentContainerConfig.getName(), topology) == null) {
                        final Container containerAttached = new Container();

                        ConfigurationServiceUtils.initializeContainer(containerAttached, currentContainerConfig);

                        topology.getDomain().getContainer().add(containerAttached);
                    }
                    // else we must be in it, and updateTopology will take care of verifying that
    
                    // The new target topology is persisted in the working area to be able to restart
                    // the container in the right topology.
                    // this will check everything is ok with the topology and update it
                    this.configurationService.updateTopology(topology);
    
                    // It is required to use IMap.put() instead of IMap.set() because IMap.set() breaks the contract
                    // of EntryListener firing an EntryEvent with a null oldValue. A non-null oldValue is required
                    // by DynamicTopologyListener.entryUpdated() to update the topology changes.
                    this.log.debug("Updating dynamic topology in shared area ...");
                    this.topologiesMap.put(this.topologyName, topology);
                    this.log.debug("Dynamic topology updated in shared area.");
    
                    // and finally activate the topology listener for future events
                    // (before releasing the lock, so that nobody can do modification before I register it!)
                    this.dynamicTopologyListener = this.topologiesMap.addEntryListener(
                            new DynamicTopologyListener(this.configurationService, this.topologyName, this.log),
                            this.topologyName, true);
                } finally {
                    this.topologyObjectLock.unlock();
                }

                reactivateAllLocalEndpoints(endpoints);

            } else {
                throw new SharedAreaException(String
                        .format("The lock on the topology object is already acquired (waited %s).", lockWaitTime));
            }
        } catch (org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException e) {
            throw new SharedAreaException("Invalid topology", e);
        } catch (InterruptedException e) {
            throw new SharedAreaException("Locking interrupted");
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    /**
     * Enable NearCache for all distributed map
     * 
     * @param cltCfg
     *            The Hazelcast client configuration
     */
    private final void enableNearCache(final ClientConfig cltCfg) {

        final ContainerConfiguration containerConfiguration = this.configurationService.getContainerConfiguration();
        final Map<String, String> extraConfProperties = containerConfiguration.getExtraConfiguration();
        int maxSize;
        if (extraConfProperties != null) {
            final String maxSizeStr = extraConfProperties.get(PROPERTY_NAME_MAP_CACHE_MAX_SIZE);
            if (maxSizeStr != null && !maxSizeStr.trim().isEmpty()) {
                try {
                    maxSize = Integer.parseInt(maxSizeStr);
                    this.log.config("Max size of map caches: " + maxSize);
                } catch (final NumberFormatException e) {
                    this.log.warning(String.format(
                            "Invalid value for the max size of map cache: '%s'. Default value used: '%s'", maxSizeStr,
                            DEFAULT_MAP_CACHE_MAX_SIZE));
                    maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
                }
            } else {
                this.log.config("The max size of map cache is not defined. Default value used: "
                        + DEFAULT_MAP_CACHE_MAX_SIZE);
                maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
            }
        } else {
            this.log.config("No extra configuration defined for the registry overlay client implementation. Default values used");
            maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
        }

        final NearCacheConfig cltNearCacheEndpointsCfg = new NearCacheConfig();
        cltNearCacheEndpointsCfg.setName(DataModelApi.getEndpointsMapName(this.topologyName));
        cltNearCacheEndpointsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointsCfg.setInvalidateOnChange(true);
        cltNearCacheEndpointsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointsCfg);

        final NearCacheConfig cltNearCacheEndpointDescriptionsCfg = new NearCacheConfig();
        cltNearCacheEndpointDescriptionsCfg.setName(DataModelApi.getEndpointDescriptionsMapName(this.topologyName));
        cltNearCacheEndpointDescriptionsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointDescriptionsCfg.setInvalidateOnChange(true);
        cltNearCacheEndpointDescriptionsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointDescriptionsCfg);

        // Note: we don't want to use near cache with the topologies because they changes a lot (w.r.t. to the use we do
        // of it) and near cache is really best for read-only things. Moreover there is overhead for checking if a value
        // must be invalidated so it's really not the best.
    }

    @Override
    public Collection<SavedEndpoint> leaveTopology() throws SharedAreaException {

        if (this.dynamicTopologyListener == null) {
            throw new SharedAreaException("Can't leave a topology if we didn't attach to one before");
        }

        // first we stop the listener because it's not relevant anymore
        // And also so that our leaving is not persisted in case other events arrive
        final boolean removed = this.topologiesMap.removeEntryListener(this.dynamicTopologyListener);
        assert removed;
        this.dynamicTopologyListener = null;

        try {
            // then we remove ourselves
            if (this.configurationService.getDomainConfiguration()
                    .getMode() == DomainConfiguration.DomainMode.DYNAMIC) {
                final int lockWaitTime = this.configurationService.getContainerConfiguration()
                        .getTopologyDynamicLockWaitTime();
                final String myName = this.configurationService.getContainerConfiguration().getName();

                if (this.topologyObjectLock.tryLock(lockWaitTime, TimeUnit.SECONDS)) {
                    try {
                        final Topology topology = this.topologiesMap.get(this.topologyName);
                        final Iterator<Container> itContainers = topology.getDomain().getContainer().iterator();
                        while (itContainers.hasNext()) {
                            if (myName.equals(itContainers.next().getName())) {
                                itContainers.remove();
                                break;
                            }
                        }
                        this.topologiesMap.put(this.topologyName, topology);
                    } finally {
                        this.topologyObjectLock.unlock();
                    }

                    return removeAllLocalEndpoints();
                } else {
                    throw new SharedAreaException(String
                            .format("The lock on the topology object is already acquired (waited %s).", lockWaitTime));
                }
            } else {
                throw new SharedAreaException("Can't leave a static topology");
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        } catch (InterruptedException e) {
            throw new SharedAreaException("Locking interrupted");
        }
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        if (this.hzlClient != null) {
            if (this.dynamicTopologyListener != null) {
                try {
                    this.topologiesMap.removeEntryListener(this.dynamicTopologyListener);
                    this.dynamicTopologyListener = null;
                } catch (final Exception e) {
                    this.log.warning("An error occured while removing the topology listener", e);
                }
            }

            try {
                this.hzlClient.shutdown();
                this.hzlClient = null;
            } catch (final Exception e) {
                this.log.warning("An error occured while stopping the registry client", e);
            }
        }

        this.log.end();
    }

    @Override
    public Properties getConfiguration(final String providedTopologyPassphrase) {
        final Properties result = new Properties();
        if (this.registryOverlayConfiguration != null) {
            final boolean isSecurityOk = this.configurationService
                    .isSecurityTopologyPassphraseOk(providedTopologyPassphrase);
            if (this.log.isDebugEnabled()) {
                if (isSecurityOk) {
                    this.log.debug("Security validated. Sensible information will be returned.");
                } else {
                    this.log.debug("Security not validated. Sensible information will NOT be returned.");
                }
            }

            result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_NAME,
                    this.registryOverlayConfiguration.getGroupName());
            if (isSecurityOk) {
                result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD,
                        this.registryOverlayConfiguration.getGroupPassword());
            }

            int addressNb = 0;
            for (final OverlayMember member : this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember()) {
                final String host = member.getValue();
                final BigInteger port = member.getPort();
                if (host != null && !host.trim().isEmpty()) {
                    result.setProperty(
                            PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS + "." + addressNb++,
                            host + ":" + (port == null ? Constants.DEFAULT_PORT : port.toString()));
                }
            }
        }
        return result;
    }

    @Override
    public Document getTopologyPartConfiguration(final Properties properties) throws MutableImplException {
        
        if (properties != null) {
            
            final Configuration xmlConfiguration = new Configuration();
    
            // Group name
            final String groupName = properties.getProperty(GROUP_NAME_PROPERTY_NAME);
            if (groupName != null) {
                xmlConfiguration.setGroupName(groupName);
            } else {
                throw new MissingPropertyException(GROUP_NAME_PROPERTY_NAME);
            }

            final String groupPwd = properties.getProperty(GROUP_PASSWORD_PROPERTY_NAME);
            if (groupPwd != null) {
                xmlConfiguration.setGroupPassword(groupPwd);
            } else {
                throw new MissingPropertyException(GROUP_PASSWORD_PROPERTY_NAME);
            }

            xmlConfiguration.setOverlayMembers(new OverlayMembers());
            boolean isAddressFound = false;
            for (final String propertyName : properties.stringPropertyNames()) {
                if (propertyName.startsWith(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS)) {
                    isAddressFound = true;
                    final String address = properties.getProperty(propertyName);
                    if (address != null && !address.isEmpty()) {
                        final String[] addressTokens = address.split(":");
                        if (addressTokens.length > 2) {
                            throw new InvalidPropertyException(propertyName, address);
                        } else {
                            final String host = addressTokens[0];
                            if (host.isEmpty()) {
                                throw new InvalidPropertyException(propertyName, address);
                            } else {
                                final OverlayMember member = new OverlayMember();
                                xmlConfiguration.getOverlayMembers().getOverlayMember().add(member);
                                member.setValue(host);

                                if (addressTokens.length == 2) {
                                    final String portStr = addressTokens[1];
                                    if (portStr != null && !portStr.isEmpty()) {
                                        member.setPort(new BigInteger(portStr));
                                    }
                                }
                            }
                        }
                    } else {
                        throw new InvalidPropertyException(propertyName, address);
                    }
                }
            }
            if (!isAddressFound) {
                throw new MissingPropertyException(MEMBER_HOST_PROPERTY_NAME_PATTERN);
            }

            try {
                final Document document = DocumentBuilders.newDocument();
                ConfigurationBuilder.marshallConfiguration(xmlConfiguration, document);
                return document;
            } catch (final ConfigurationException e) {
                throw new MutableImplException(e);
            }
        } else {
            return null;
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForService(final QName serviceName, final EndpointType type)
            throws SharedAreaException {

        if (serviceName == null) {
            throw new SharedAreaException("No serviceName found");
        }

        try {
            // values() returns a copy of the values
            final SqlPredicate snP = new SqlPredicate(
                    String.format("serviceName = '%s'", serviceName.toString()));
            final Collection<StoredRegistryEndpoint> values = this.endpointsMap
                    .values(type == null ? snP : Predicates.and(Predicates.equal("type", toStoredType(type)), snP));
            return Collections2.transform(values, STORED_2_PETALS);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public Document retrieveEndpointDescription(final String endpointName) throws SharedAreaException {
        try {
            return toDocument(this.endpointDescriptionsMap.get(endpointName));
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    private static Document toDocument(final String desc) throws SharedAreaException {
        try {
            return desc != null ? XMLHelper.createDocumentFromString(desc) : null;
        } catch (final IOException | SAXException e) {
            throw new SharedAreaException("Can't parse the descriptions stored in the registry", e);
        }
    }
    
    @Override
    public Document getDescription(final ServiceEndpoint endpoint) throws SharedAreaException {
        return retrieveEndpointDescription(endpoint.getEndpointName());
    }

    @Override
    public Document getDescription(final QName serviceName, final String endpointName) throws SharedAreaException {
        return retrieveEndpointDescription(endpointName);
    }

    @Override
    public String getImplementationName() {
        return "Petals Distributed Shared Memory Overlay Client";
    }

    @Override
    public String getImplementationVersion() {
        return this.getClass().getPackage().getImplementationVersion();
    }

    private final static Function<StoredRegistryEndpoint, PetalsServiceEndpoint> STORED_2_PETALS = new Function<StoredRegistryEndpoint, PetalsServiceEndpoint>() {
        @Override
        public PetalsServiceEndpoint apply(final StoredRegistryEndpoint input) {
            switch (input.getType()) {
                case INTERNAL:
                    return new InternalEndpointImpl(input);
                case EXTERNAL:
                    return new ExternalEndpointImpl(input);
            }
            throw new IllegalArgumentException("impossible case");
        }
    };

    private static Type toStoredType(final EndpointType type) {
        switch (type) {
            case EXTERNAL:
                return Type.EXTERNAL;
            case INTERNAL:
                return Type.INTERNAL;
        }
        throw new IllegalArgumentException("impossible case");
    }

    private static StoredRegistryEndpoint toStoredEndpoint(final Type type, final ServiceEndpoint endpoint,
            final Location location) {
        return toStoredEndpoint(type, endpoint.getEndpointName(), endpoint.getServiceName(), endpoint.getInterfaces(),
                location);
    }

    private static StoredRegistryEndpoint toStoredEndpoint(final Type type, final String endpointName,
            final QName serviceName, final QName[] interfaces, final Location location) {
        final int l = interfaces.length;
        final String[] interfacesString = new String[l];
        for (int i = 0; i < l; i++) {
            interfacesString[i] = interfaces[i].toString();
        }
        return new StoredRegistryEndpoint(type, endpointName, serviceName.toString(), interfacesString,
                location.getContainerName(), location.getComponentName());
    }
}
